from sigma.conversion.state import ConversionState
from sigma.processing.pipeline import ProcessingPipeline
from sigma.rule import SigmaRule
from sigma.conversion.base import TextQueryBackend
from sigma.conversion.deferred import DeferredQueryExpression
from sigma.conditions import ConditionItem, ConditionAND, ConditionOR, ConditionNOT, ConditionFieldEqualsValueExpression
from sigma.types import SigmaCompareExpression, SigmaString
from sigma.pipelines.microsoft365defender import microsoft_365_defender_pipeline
import sigma
import re
from typing import ClassVar, Dict, Tuple, Pattern, Any, Union


class Microsoft365DefenderBackend(TextQueryBackend):
    """Microsoft 365 Defender KQL Backend"""

    # Automatically apply pipeline
    backend_processing_pipeline: ClassVar[ProcessingPipeline] = microsoft_365_defender_pipeline()

    # The backend generates grouping if required
    name: ClassVar[str] = "Microsoft 365 Defender backend"
    formats: Dict[str, str] = {
        "default": "KQL for Microsoft 365 Defender Advanced Hunting queries",
    }

    requires_pipeline: bool = False  # m365 pipeline is automatically applied

    # Operator precedence
    precedence: ClassVar[Tuple[ConditionItem, ConditionItem, ConditionItem]] = (ConditionNOT, ConditionOR, ConditionAND)
    group_expression: ClassVar[
        str] = "({expr})"  # Expression for precedence override grouping as format string with {expr} placeholder

    # Generated query tokens
    token_separator: str = " "  # separator inserted between all boolean operators
    or_token: ClassVar[str] = "or"
    and_token: ClassVar[str] = "and"
    not_token: ClassVar[str] = " !~ "
    eq_token: ClassVar[str] = " =~ "  # Token inserted between field and value (without separator)

    # String output
    ## Fields
    ### Quoting
    field_quote: ClassVar[
        str] = "'"  # Character used to quote field characters if field_quote_pattern matches (or not, depending on field_quote_pattern_negation). No field name quoting is done if not set.
    field_quote_pattern: ClassVar[Pattern] = re.compile(
        "^\\w+$")  # Quote field names if this pattern (doesn't) matches, depending on field_quote_pattern_negation. Field name is always quoted if pattern is not set.
    field_quote_pattern_negation: ClassVar[
        bool] = True  # Negate field_quote_pattern result. Field name is quoted if pattern doesn't matches if set to True (default).

    ### Escaping
    field_escape: ClassVar[str] = ""  # Character to escape particular parts defined in field_escape_pattern.
    field_escape_quote: ClassVar[bool] = True  # Escape quote string defined in field_quote
    field_escape_pattern: ClassVar[Pattern] = re.compile(
        "\\s")  # All matches of this pattern are prepended with the string contained in field_escape.

    ## Values
    str_quote: ClassVar[str] = '"'  # string quoting character (added as escaping character)
    escape_char: ClassVar[str] = "\\"  # Escaping character for special characrers inside string
    wildcard_multi: ClassVar[str] = "*"  # Character used as multi-character wildcard
    wildcard_single: ClassVar[str] = "*"  # Character used as single-character wildcard
    add_escaped: ClassVar[str] = "\\"  # Characters quoted in addition to wildcards and string quote
    filter_chars: ClassVar[str] = ""  # Characters filtered
    bool_values: ClassVar[Dict[bool, str]] = {  # Values to which boolean values are mapped.
        True: "true",
        False: "false",
    }

    # String matching operators. if none is appropriate eq_token is used.
    startswith_expression: ClassVar[str] = "{field} startswith {value}"
    endswith_expression: ClassVar[str] = "{field} endswith {value}"
    contains_expression: ClassVar[str] = "{field} contains {value}"
    wildcard_match_expression: ClassVar[
        str] = "match"  # Special expression if wildcards can't be matched with the eq_token operator

    # Regular expressions
    re_expression: ClassVar[
        str] = "{field} matches regex \"{regex}\""  # Regular expression query as format string with placeholders {field} and {regex}
    re_escape_char: ClassVar[str] = "\\"  # Character used for escaping in regular expressions
    re_escape: ClassVar[Tuple[str]] = ()  # List of strings that are escaped
    re_escape_escape_char: bool = True  # If True, the escape character is also escaped

    # cidr expressions
    cidr_wildcard: ClassVar[str] = "*"  # Character used as single wildcard
    cidr_expression: ClassVar[
        str] = "ipv4_is_in_range({field}, \"{value}\")"  # CIDR expression query as format string with placeholders {field} = {value}
    cidr_in_list_expression: ClassVar[
        str] = "ipv4_is_in_any_range({field}, \"{value}\")"  # CIDR expression query as format string with placeholders {field} = in({list})

    # Numeric comparison operators
    compare_op_expression: ClassVar[
        str] = "{field} {operator} {value}"  # Compare operation query as format string with placeholders {field}, {operator} and {value}
    # Mapping between CompareOperators elements and strings used as replacement for {operator} in compare_op_expression
    compare_operators: ClassVar[Dict[SigmaCompareExpression.CompareOperators, str]] = {
        SigmaCompareExpression.CompareOperators.LT: "<",
        SigmaCompareExpression.CompareOperators.LTE: "<=",
        SigmaCompareExpression.CompareOperators.GT: ">",
        SigmaCompareExpression.CompareOperators.GTE: ">=",
    }

    # Null/None expressions
    field_null_expression: ClassVar[
        str] = "isnull({field})"  # Expression for field has null value as format string with {field} placeholder for field name

    # Field value in list, e.g. "field in (value list)" or "field containsall (value list)"
    convert_or_as_in: ClassVar[bool] = True  # Convert OR as in-expression
    convert_and_as_in: ClassVar[bool] = True  # Convert AND as in-expression
    in_expressions_allow_wildcards: ClassVar[
        bool] = True  # Values in list can contain wildcards. If set to False (default) only plain values are converted into in-expressions.
    field_in_list_expression: ClassVar[
        str] = "{field} {op} ({list})"  # Expression for field in list of values as format string with placeholders {field}, {op} and {list}
    or_in_operator: ClassVar[
        str] = "has_any"  # Operator used to convert OR into in-expressions. Must be set if convert_or_as_in is set
    and_in_operator: ClassVar[
        str] = "has_all"  # Operator used to convert AND into in-expressions. Must be set if convert_and_as_in is set
    list_separator: ClassVar[str] = ", "  # List element separator

    # Value not bound to a field
    unbound_value_str_expression: ClassVar[
        str] = '{value}'  # Expression for string value not bound to a field as format string with placeholder {value}
    unbound_value_num_expression: ClassVar[
        str] = '{value}'  # Expression for number value not bound to a field as format string with placeholder {value}
    unbound_value_re_expression: ClassVar[
        str] = '_=~{value}'  # Expression for regular expression not bound to a field as format string with placeholder {value}

    # Query finalization: appending and concatenating deferred query part
    deferred_start: ClassVar[str] = "\n| "  # String used as separator between main query and deferred parts
    deferred_separator: ClassVar[str] = "\n| "  # String used to join multiple deferred query parts
    deferred_only_query: ClassVar[str] = "*"  # String used as query if final query only contains deferred expression

    # TODO: implement custom methods for query elements not covered by the default backend base.
    # Documentation: https://sigmahq-pysigma.readthedocs.io/en/latest/Backends.html

    # We use =~ for eq_token so everything is case insensitive. But this cannot be used with ints/numbers in queries
    # So we can define a new token to use for SigmaNumeric types and override convert_condition_field_eq_val_num
    # to use it
    num_eq_token: ClassVar[str] = " == "

    # Override methods
    #  For numeric values, need == instead of =~
    def convert_condition_field_eq_val_num(self, cond: ConditionFieldEqualsValueExpression, state: ConversionState) -> \
            Union[str, DeferredQueryExpression]:
        """Conversion of field = number value expressions"""
        try:
            return self.escape_and_quote_field(cond.field) + self.num_eq_token + str(cond.value)
        except TypeError:  # pragma: no cover
            raise NotImplementedError("Field equals numeric value expressions are not supported by the backend.")

    def convert_condition_as_in_expression(self,
                                           cond: Union[ConditionOR, ConditionAND],
                                           state: ConversionState) -> Union[str, DeferredQueryExpression]:
        """Overridden method for conversion of field in value list conditions.
        KQL doesn't really use wildcards, so if we have an 'as_in' condition where one or more of the values has a wildcard,
        we can still use the as_in condition, then append on the wildcard value(s) with a startswith, endswith, or contains
        expression
        """

        field = self.escape_and_quote_field(cond.args[0].field)
        op1 = self.or_in_operator if isinstance(cond, ConditionOR) else self.and_in_operator
        op2 = self.or_token if isinstance(cond, ConditionOR) else self.and_token
        list_nonwildcard = self.list_separator.join([
            self.convert_value_str(arg.value, state)
            for arg in cond.args
            if (isinstance(arg.value, SigmaString) and not arg.value.contains_special()) or not isinstance(arg.value,
                                                                                                           SigmaString)
        ])
        list_wildcards = [
            arg.value for arg in cond.args if isinstance(arg.value, SigmaString) and arg.value.contains_special()
        ]
        as_in_expr = ""
        # Convert as_in and wildcard values separately
        if list_nonwildcard:
            as_in_expr = self.field_in_list_expression.format(
                field=field,
                op=op1,
                list=list_nonwildcard
            )
        wildcard_exprs_list = []
        if list_wildcards:
            for arg in list_wildcards:
                new_cond = ConditionFieldEqualsValueExpression(field=field, value=arg)
                expr = self.convert_condition_field_eq_val_str(new_cond, state)
                wildcard_exprs_list.append(expr)
        wildcard_exprs = f'{self.token_separator}{op2}{self.token_separator}'.join(wildcard_exprs_list)
        if as_in_expr and wildcard_exprs:
            return as_in_expr + self.token_separator + op2 + self.token_separator + wildcard_exprs
        return as_in_expr + wildcard_exprs

    def finalize_query_default(self, rule: SigmaRule, query: Any, index: int, state: ConversionState) -> Any:
        """
        Finalize conversion result of a query. The classes default behavior is to just return the query.
        We will add the table name to the beginning of the query based on the category of the rule.
        This is taken from the ProcessingPipeline state, which is why we automatically run the
        pipeline as part of the backend.
        """
        query_table = state.processing_state.get('query_table', None)
        query_table = query_table + "\n| where " if query_table else "search "

        return query_table + query



from .microsoft365defender import microsoft_365_defender_pipeline
# TODO: add all pipelines that should be exposed to the user of your backend in the import statement above.

pipelines = {
    "microsoft_365_defender_pipeline": microsoft_365_defender_pipeline,   # TODO: adapt identifier to something approproiate
}


from typing import Union, Optional, Iterable

from sigma.exceptions import SigmaTransformationError
from sigma.pipelines.common import logsource_windows_process_creation, \
    logsource_windows_image_load, logsource_windows_file_event, logsource_windows_file_delete, \
    logsource_windows_file_change, logsource_windows_file_access, logsource_windows_file_rename, \
    logsource_windows_registry_set, logsource_windows_registry_add, logsource_windows_registry_delete, \
    logsource_windows_registry_event, logsource_windows_network_connection
from sigma.processing.transformations import FieldMappingTransformation, \
    RuleFailureTransformation, ReplaceStringTransformation, SetStateTransformation, DetectionItemTransformation, \
    ValueTransformation, DetectionItemFailureTransformation
from sigma.processing.conditions import IncludeFieldCondition, \
    RuleProcessingItemAppliedCondition, ExcludeFieldCondition, DetectionItemProcessingItemAppliedCondition
from sigma.conditions import ConditionOR
from sigma.types import SigmaString, SigmaType
from sigma.processing.pipeline import ProcessingItem, ProcessingPipeline
from sigma.rule import SigmaDetectionItem, SigmaDetection
from collections import defaultdict


# CUSTOM TRANSFORMATIONS
## Custom DetectionItemTransformation to split domain and user, if applicable
class SplitDomainUserTransformation(DetectionItemTransformation):
    """Custom DetectionItemTransformation transformation to split a User field into separate domain and user fields,
    if applicable.  This is to handle the case where the Sysmon `User` field may contain a domain AND username, and
    Advanced Hunting queries separate out the domain and username into separate fields.
    If a matching field_name_condition field uses the schema DOMAIN\\USER, a new SigmaDetectionItem
    will be made for the Domain and put inside a SigmaDetection with the original User SigmaDetectionItem
    (minus the domain) for the matching SigmaDetectionItem.

    You should use this with a field_name_condition for `IncludeFieldName(['field', 'names', 'for', 'username']`)"""

    def apply_detection_item(self, detection_item: SigmaDetectionItem) -> Optional[
        Union[SigmaDetection, SigmaDetectionItem]]:
        to_return = []
        if not isinstance(detection_item.value, list):  # Ensure its a list, but it most likely will be
            detection_item.value = list(detection_item.value)
        for d in detection_item.value:
            username = d.to_plain().split("\\")
            username_field_mappings = {
                'AccountName': 'AccountDomain',
                'RequestAccountName': 'RequestAccountDomain',
                'InitiatingProcessAccountName': 'InitiatingProcessAccountDomain',
            }
            if len(username) == 2:
                domain = username[0]
                username = [SigmaString(username[1])]

                domain_field = username_field_mappings.get(detection_item.field, "InitiatingProcessAccountDomain")
                domain_value = [SigmaString(domain)]
                user_detection_item = SigmaDetectionItem(field=detection_item.field,
                                                         modifiers=[],
                                                         value=username,
                                                         )
                domain_detection_item = SigmaDetectionItem(field=domain_field,
                                                           modifiers=[],
                                                           value=domain_value)
                to_return.append(SigmaDetection(detection_items=[user_detection_item, domain_detection_item]))
            else:

                to_return.append(SigmaDetection([SigmaDetectionItem(field=detection_item.field,
                                                                    modifiers=detection_item.modifiers,
                                                                    value=username)]))
        return SigmaDetection(to_return)


## Custom DetectionItemTransformation to regex hash algos/values in Hashes field, if applicable
class HashesValuesTransformation(DetectionItemTransformation):
    """Custom DetectionItemTransformation to take a list of values in the 'Hashes' field, which are expected to be
    'algo:hash_value', and create new SigmaDetectionItems for each hash type, where the values is a list of
    SigmaString hashes. If the hash type is not part of the value, it will be inferred based on length.

    Use with field_name_condition for Hashes field"""

    def apply_detection_item(self, detection_item: SigmaDetectionItem) -> Optional[
        Union[SigmaDetection, SigmaDetectionItem]]:
        to_return = []
        algo_dict = defaultdict(list)  # map to keep track of algos and lists of values
        if not isinstance(detection_item.value, list):
            detection_item.value = [detection_item.value]
        for d in detection_item.value:
            hash_value = d.to_plain().split("|")
            if len(hash_value) == 2:
                hash_algo = hash_value[0].upper() if hash_value[0].upper() in ['MD5', 'SHA1', 'SHA256'] else ""
                hash_value = hash_value[1]
            else:
                hash_value = hash_value[0]
                if len(hash_value) == 32:  # MD5
                    hash_algo = 'MD5'
                elif len(hash_value) == 40:  # SHA1
                    hash_algo = 'SHA1'
                elif len(hash_value) == 64:  # SHA256
                    hash_algo = "SHA256"
                else:  # Invalid algo, no fieldname for keyword search
                    hash_algo = ''
            algo_dict[hash_algo].append(hash_value)
        for k, v in algo_dict.items():
            if k:  # Filter out invalid hash algo types
                to_return.append(SigmaDetectionItem(field=k if k != 'keyword' else None,
                                                    modifiers=[],
                                                    value=[SigmaString(x) for x in v]))
        return SigmaDetection(detection_items=to_return, item_linking=ConditionOR)


## Change ActionType value AFTER field transformations from Sysmon values to DeviceRegistryEvents values
class RegistryActionTypeValueTransformation(ValueTransformation):
    """Custom ValueTransformation transformation. The Microsoft DeviceRegistryEvents table expect the ActionType to
    be a slightly different set of values than what Sysmon specified, so this will change them to the correct value."""
    value_mappings = {  # Sysmon EventType -> DeviceRegistyEvents ActionType
        'CreateKey': 'RegistryKeyCreated',
        'DeleteKey': ['RegistryKeyDeleted', 'RegistryValueDeleted'],
        'SetValue': 'RegistryValueSet',
        'RenameKey': ['RegistryValueSet', 'RegistryKeyCreated'],
    }

    def apply_value(self, field: str, val: SigmaType) -> Optional[Union[SigmaType, Iterable[SigmaType]]]:
        mapped_vals = self.value_mappings.get(val.to_plain(), val.to_plain())
        if isinstance(mapped_vals, list):
            return [SigmaString(v) for v in mapped_vals]
        return SigmaString(mapped_vals)


class InvalidFieldTransformation(DetectionItemFailureTransformation):
    """
    Overrides the apply_detection_item() method from DetectionItemFailureTransformation to also include the field name
    in the error message
    """

    def apply_detection_item(self, detection_item: SigmaDetectionItem) -> None:
        field_name = detection_item.field
        self.message = f"Invalid SigmaDetectionItem field name encountered: {field_name}. " + self.message
        raise SigmaTransformationError(self.message)


# FIELD MAPPINGS
## Field mappings from Sysmon (where applicable) fields to Advanced Hunting Query fields based on schema in tables
## See: https://learn.microsoft.com/en-us/microsoft-365/security/defender/advanced-hunting-schema-tables?view=o365-worldwide#learn-the-schema-tables
query_table_field_mappings = {
    'DeviceProcessEvents': {  # process_creation, Sysmon EventID 1 -> DeviceProcessEvents table
        # ProcessGuid: ?,
        'ProcessId': 'ProcessId',
        'Image': 'FolderPath',
        'FileVersion': 'ProcessVersionInfoProductVersion',
        'Description': 'ProcessVersionInfoFileDescription',
        'Product': 'ProcessVersionInfoProductName',
        'Company': 'ProcessVersionInfoCompanyName',
        'OriginalFileName': 'ProcessVersionInfoOriginalFileName',
        'CommandLine': 'ProcessCommandLine',
        # CurrentDirectory: ?
        'User': 'AccountName',
        # LogonGuid: ?
        'LogonId': 'LogonId',
        # TerminalSessionId: ?
        'IntegrityLevel': 'ProcessIntegrityLevel',
        'sha1': 'SHA1',
        'sha256': 'SHA256',
        'md5': 'MD5',
        # 'ParentProcessGuid': ?,
        'ParentProcessId': 'InitiatingProcessId',
        'ParentImage': 'InitiatingProcessFolderPath',
        'ParentCommandLine': 'InitiatingProcessCommandLine',
        'ParentUser': 'InitiatingProcessAccountName',
    },
    'DeviceImageLoadEvents': {
        # 'ProcessGuid': ?,
        'ProcessId': 'InitiatingProcessId',
        'Image': 'InitiatingProcessFolderPath',  # File path of the process that loaded the image
        'ImageLoaded': 'FolderPath',
        'FileVersion': 'InitiatingProcessVersionInfoProductVersion',
        'Description': 'InitiatingProcessVersionInfoFileDescription',
        'Product': 'InitiatingProcessVersionInfoProductName',
        'Company': 'InitiatingProcessVersionInfoCompanyName',
        'OriginalFileName': 'InitiatingProcessVersionInfoOriginalFileName',
        # 'Hashes': ?,
        'sha1': 'SHA1',
        'sha256': 'SHA256',
        'md5': 'MD5',
        # 'Signed': ?
        # 'Signature': ?
        # 'SignatureStatus': ?
        'User': 'InitiatingProcessAccountName'
    },
    'DeviceFileEvents': {  # file_*, Sysmon EventID 11 (create), 23 (delete) -> DeviceFileEvents table
        # 'ProcessGuid': ?,
        'ProcessId': 'InitiatingProcessId',
        'Image': 'InitiatingProcessFolderPath',
        'TargetFilename': 'FolderPath',
        # 'CreationUtcTime': 'Timestamp',
        'User': 'RequestAccountName',
        # 'Hashes': ?,
        'sha1': 'SHA1',
        'sha256': 'SHA256',
        'md5': 'MD5',
    },
    'DeviceNetworkEvents': {  # network_connection, Sysmon EventID 3 -> DeviceNetworkEvents table
        # 'ProcessGuid': ?,
        'ProcessId': 'InitiatingProcessId',
        'Image': 'InitiatingProcessFolderPath',
        'User': 'InitiatingProcessAccountName',
        'Protocol': 'Protocol',
        # 'Initiated': ?,
        # 'SourceIsIpv6': ?,
        'SourceIp': 'LocalIP',
        'SourceHostname': 'DeviceName',
        'SourcePort': 'LocalPort',
        # 'SourcePortName': ?,
        # 'DestinationIsIpv6': ?,
        'DestinationIp': 'RemoteIP',
        'DestinationHostname': 'RemoteUrl',
        'DestinationPort': 'RemotePort',
        # 'DestinationPortName': ?,
    },
    "DeviceRegistryEvents": {
        # registry_*, Sysmon EventID 12 (create/delete), 13 (value set), 14 (key/value rename) -> DeviceRegistryEvents table,
        'EventType': 'ActionType',
        # 'ProcessGuid': ?,
        'ProcessId': 'InitiatingProcessId',
        'Image': 'InitiatingProcessFolderPath',
        'TargetObject': 'RegistryKey',
        # 'NewName': ?
        'Details': 'RegistryValueData',
        'User': 'InitiatingProcessAccountName'
    }
}

## Generic catch-all field mappings for sysmon -> microsoft 365 defender fields that appear in most tables and
## haven't been mapped already
generic_field_mappings = {
    'EventType': 'ActionType',
    'User': 'InitiatingProcessAccountName',
    'CommandLine': 'InitiatingProcessCommandLine',
    'Image': 'InitiatingProcessFolderPath',
    'SourceImage': 'InitiatingProcessFolderPath',
    'ProcessId': 'InitiatingProcessId',
    'md5': 'InitiatingProcessMD5',
    'sha1': 'InitiatingProcessSHA1',
    'sha256': 'InitiatingProcessSHA256',
    'ParentProcessId': 'InitiatingProcessParentId',
    'ParentCommandLine': 'InitiatingProcessParentCommandLine',
    'Company': 'InitiatingProcessVersionInfoCompanyName',
    'Description': 'InitiatingProcessVersionInfoFileDescription',
    'OriginalFileName': 'InitiatingProcessVersionInfoOriginalFileName',
    'Product': 'InitiatingProcessVersionInfoProductName'
}

# VALID FIELDS PER QUERY TABLE
## Will Implement field checking later once issue with removing fields is figured out, for now it fails the pipeline
## dict of {'table_name': [list, of, valid_fields]} for each table
valid_fields_per_table = {
    'DeviceProcessEvents': ['Timestamp', 'DeviceId', 'DeviceName', 'ActionType', 'FileName', 'FolderPath', 'SHA1',
                            'SHA256', 'MD5', 'FileSize', 'ProcessVersionInfoCompanyName',
                            'ProcessVersionInfoProductName', 'ProcessVersionInfoProductVersion',
                            'ProcessVersionInfoInternalFileName', 'ProcessVersionInfoOriginalFileName',
                            'ProcessVersionInfoFileDescription', 'ProcessId', 'ProcessCommandLine',
                            'ProcessIntegrityLevel', 'ProcessTokenElevation', 'ProcessCreationTime', 'AccountDomain',
                            'AccountName', 'AccountSid', 'AccountUpn', 'AccountObjectId', 'LogonId',
                            'InitiatingProcessAccountDomain', 'InitiatingProcessAccountName',
                            'InitiatingProcessAccountSid', 'InitiatingProcessAccountUpn',
                            'InitiatingProcessAccountObjectId', 'InitiatingProcessLogonId',
                            'InitiatingProcessIntegrityLevel', 'InitiatingProcessTokenElevation',
                            'InitiatingProcessSHA1', 'InitiatingProcessSHA256', 'InitiatingProcessMD5',
                            'InitiatingProcessFileName', 'InitiatingProcessFileSize',
                            'InitiatingProcessVersionInfoCompanyName', 'InitiatingProcessVersionInfoProductName',
                            'InitiatingProcessVersionInfoProductVersion',
                            'InitiatingProcessVersionInfoInternalFileName',
                            'InitiatingProcessVersionInfoOriginalFileName',
                            'InitiatingProcessVersionInfoFileDescription', 'InitiatingProcessId',
                            'InitiatingProcessCommandLine', 'InitiatingProcessCreationTime',
                            'InitiatingProcessFolderPath', 'InitiatingProcessParentId',
                            'InitiatingProcessParentFileName', 'InitiatingProcessParentCreationTime',
                            'InitiatingProcessSignerType', 'InitiatingProcessSignatureStatus', 'ReportId',
                            'AppGuardContainerId', 'AdditionalFields'],
    'DeviceImageLoadEvents': ['Timestamp', 'DeviceId', 'DeviceName', 'ActionType', 'FileName', 'FolderPath', 'SHA1',
                              'SHA256', 'MD5', 'FileSize', 'InitiatingProcessAccountDomain',
                              'InitiatingProcessAccountName', 'InitiatingProcessAccountSid',
                              'InitiatingProcessAccountUpn', 'InitiatingProcessAccountObjectId',
                              'InitiatingProcessIntegrityLevel', 'InitiatingProcessTokenElevation',
                              'InitiatingProcessSHA1', 'InitiatingProcessSHA256', 'InitiatingProcessMD5',
                              'InitiatingProcessFileName', 'InitiatingProcessFileSize',
                              'InitiatingProcessVersionInfoCompanyName', 'InitiatingProcessVersionInfoProductName',
                              'InitiatingProcessVersionInfoProductVersion',
                              'InitiatingProcessVersionInfoInternalFileName',
                              'InitiatingProcessVersionInfoOriginalFileName',
                              'InitiatingProcessVersionInfoFileDescription', 'InitiatingProcessId',
                              'InitiatingProcessCommandLine', 'InitiatingProcessCreationTime',
                              'InitiatingProcessFolderPath', 'InitiatingProcessParentId',
                              'InitiatingProcessParentFileName', 'InitiatingProcessParentCreationTime', 'ReportId',
                              'AppGuardContainerId'],
    'DeviceFileEvents': ['Timestamp', 'DeviceId', 'DeviceName', 'ActionType', 'FileName', 'FolderPath', 'SHA1',
                         'SHA256', 'MD5', 'FileOriginUrl', 'FileOriginReferrerUrl', 'FileOriginIP',
                         'PreviousFolderPath', 'PreviousFileName', 'FileSize', 'InitiatingProcessAccountDomain',
                         'InitiatingProcessAccountName', 'InitiatingProcessAccountSid', 'InitiatingProcessAccountUpn',
                         'InitiatingProcessAccountObjectId', 'InitiatingProcessMD5', 'InitiatingProcessSHA1',
                         'InitiatingProcessSHA256', 'InitiatingProcessFolderPath', 'InitiatingProcessFileName',
                         'InitiatingProcessFileSize', 'InitiatingProcessVersionInfoCompanyName',
                         'InitiatingProcessVersionInfoProductName', 'InitiatingProcessVersionInfoProductVersion',
                         'InitiatingProcessVersionInfoInternalFileName', 'InitiatingProcessVersionInfoOriginalFileName',
                         'InitiatingProcessVersionInfoFileDescription', 'InitiatingProcessId',
                         'InitiatingProcessCommandLine', 'InitiatingProcessCreationTime',
                         'InitiatingProcessIntegrityLevel', 'InitiatingProcessTokenElevation',
                         'InitiatingProcessParentId', 'InitiatingProcessParentFileName',
                         'InitiatingProcessParentCreationTime', 'RequestProtocol', 'RequestSourceIP',
                         'RequestSourcePort', 'RequestAccountName', 'RequestAccountDomain', 'RequestAccountSid',
                         'ShareName', 'InitiatingProcessFileSize', 'SensitivityLabel', 'SensitivitySubLabel',
                         'IsAzureInfoProtectionApplied', 'ReportId', 'AppGuardContainerId', 'AdditionalFields'],
    'DeviceRegistryEvents': ['Timestamp', 'DeviceId', 'DeviceName', 'ActionType', 'RegistryKey', 'RegistryValueType',
                             'RegistryValueName', 'RegistryValueData', 'PreviousRegistryKey',
                             'PreviousRegistryValueName', 'PreviousRegistryValueData', 'InitiatingProcessAccountDomain',
                             'InitiatingProcessAccountName', 'InitiatingProcessAccountSid',
                             'InitiatingProcessAccountUpn', 'InitiatingProcessAccountObjectId', 'InitiatingProcessSHA1',
                             'InitiatingProcessSHA256', 'InitiatingProcessMD5', 'InitiatingProcessFileName',
                             'InitiatingProcessFileSize', 'InitiatingProcessVersionInfoCompanyName',
                             'InitiatingProcessVersionInfoProductName', 'InitiatingProcessVersionInfoProductVersion',
                             'InitiatingProcessVersionInfoInternalFileName',
                             'InitiatingProcessVersionInfoOriginalFileName',
                             'InitiatingProcessVersionInfoFileDescription', 'InitiatingProcessId',
                             'InitiatingProcessCommandLine', 'InitiatingProcessCreationTime',
                             'InitiatingProcessFolderPath', 'InitiatingProcessParentId',
                             'InitiatingProcessParentFileName', 'InitiatingProcessParentCreationTime',
                             'InitiatingProcessIntegrityLevel', 'InitiatingProcessTokenElevation', 'ReportId',
                             'AppGuardContainerId'],
    'DeviceNetworkEvents': ['Timestamp', 'DeviceId', 'DeviceName', 'ActionType', 'RemoteIP', 'RemotePort', 'RemoteUrl',
                            'LocalIP', 'LocalPort', 'Protocol', 'LocalIPType', 'RemoteIPType', 'InitiatingProcessSHA1',
                            'InitiatingProcessSHA256', 'InitiatingProcessMD5', 'InitiatingProcessFileName',
                            'InitiatingProcessFileSize', 'InitiatingProcessVersionInfoCompanyName',
                            'InitiatingProcessVersionInfoProductName', 'InitiatingProcessVersionInfoProductVersion',
                            'InitiatingProcessVersionInfoInternalFileName',
                            'InitiatingProcessVersionInfoOriginalFileName',
                            'InitiatingProcessVersionInfoFileDescription', 'InitiatingProcessId',
                            'InitiatingProcessCommandLine', 'InitiatingProcessCreationTime',
                            'InitiatingProcessFolderPath', 'InitiatingProcessParentFileName',
                            'InitiatingProcessParentId', 'InitiatingProcessParentCreationTime',
                            'InitiatingProcessAccountDomain', 'InitiatingProcessAccountName',
                            'InitiatingProcessAccountSid', 'InitiatingProcessAccountUpn',
                            'InitiatingProcessAccountObjectId', 'InitiatingProcessIntegrityLevel',
                            'InitiatingProcessTokenElevation', 'ReportId', 'AppGuardContainerId', 'AdditionalFields']}

# OTHER MAPPINGS
## useful for creating ProcessingItems() with list comprehension

## Query Table names -> rule categories
table_to_category_mappings = {
    'DeviceProcessEvents': ['process_creation'],
    'DeviceImageLoadEvents': ['image_load'],
    'DeviceFileEvents': ['file_access', 'file_change', 'file_delete', 'file_event', 'file_rename'],
    'DeviceRegistryEvents': ['registry_add', 'registry_delete', 'registry_event', 'registry_set'],
    'DeviceNetworkEvents': ['network_connection']
}

## rule categories -> RuleConditions
category_to_conditions_mappings = {
    'process_creation': logsource_windows_process_creation(),
    'image_load': logsource_windows_image_load(),
    'file_access': logsource_windows_file_access(),
    'file_change': logsource_windows_file_change(),
    'file_delete': logsource_windows_file_delete(),
    'file_event': logsource_windows_file_event(),
    'file_rename': logsource_windows_file_rename(),
    'registry_add': logsource_windows_registry_add(),
    'registry_delete': logsource_windows_registry_delete(),
    'registry_event': logsource_windows_registry_event(),
    'registry_set': logsource_windows_registry_set(),
    'network_connection': logsource_windows_network_connection()
}

# PROCESSING_ITEMS()
## ProcessingItems to set state key 'query_table' to use in backend
## i.e. $QueryTable$ | $rest_of_query$
query_table_proc_items = [
    ProcessingItem(
        identifier=f"microsoft_365_defender_set_query_table_{table_name}",
        transformation=SetStateTransformation("query_table", table_name),
        rule_conditions=[
            category_to_conditions_mappings[rule_category] for rule_category in rule_categories
        ],
        rule_condition_linking=any,
    )
    for table_name, rule_categories in table_to_category_mappings.items()
]

## Fieldmappings
fieldmappings_proc_items = [
    ProcessingItem(
        identifier=f"microsoft_365_defender_fieldmappings_{table_name}",
        transformation=FieldMappingTransformation(query_table_field_mappings[table_name]),
        rule_conditions=[
            category_to_conditions_mappings[rule_category] for rule_category in rule_categories
        ],
        rule_condition_linking=any,
    )
    for table_name, rule_categories in table_to_category_mappings.items()
]

## Generic Fielp Mappings, keep this last
## Exclude any fields already mapped. For example, if process_creation events ProcessId has already
## been mapped to the same field name (ProcessId), we don't to remap it to InitiatingProcessId
generic_field_mappings_proc_item = [ProcessingItem(
    identifier="microsoft_365_defender_fieldmappings_generic",
    transformation=FieldMappingTransformation(
        generic_field_mappings
    ),
    detection_item_conditions=[
        DetectionItemProcessingItemAppliedCondition(f"microsoft_365_defender_fieldmappings_{table_name}")
        for table_name in table_to_category_mappings.keys()
    ],
    detection_item_condition_linking=any,
    detection_item_condition_negation=True,
)
]
## Field Value Replacements ProcessingItems
replacement_proc_items = [
    # Sysmon uses abbreviations in RegistryKey values, replace with full key names as the DeviceRegistryEvents schema
    # expects them to be
    # Note: Ensure this comes AFTER field mapping renames, as we're specifying DeviceRegistryEvent fields
    #
    # Do this one first, or else the HKLM only one will replace HKLM and mess up the regex
    ProcessingItem(
        identifier="microsoft_365_defender_registry_key_replace_currentcontrolset",
        transformation=ReplaceStringTransformation(regex=r"(?i)(^HKLM\\SYSTEM\\CurrentControlSet)",
                                                   replacement=r"HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet001"),
        field_name_conditions=[IncludeFieldCondition(['RegistryKey', 'PreviousRegistryKey'])]
    ),
    ProcessingItem(
        identifier="microsoft_365_defender_registry_key_replace_hklm",
        transformation=ReplaceStringTransformation(regex=r"(?i)(^HKLM)",
                                                   replacement=r"HKEY_LOCAL_MACHINE"),
        field_name_conditions=[IncludeFieldCondition(['RegistryKey', 'PreviousRegistryKey'])]
    ),
    ProcessingItem(
        identifier="microsoft_365_defender_registry_key_replace_hku",
        transformation=ReplaceStringTransformation(regex=r"(?i)(^HKU)",
                                                   replacement=r"HKEY_USERS"),
        field_name_conditions=[IncludeFieldCondition(['RegistryKey', 'PreviousRegistryKey'])]
    ),
    ProcessingItem(
        identifier="microsoft_365_defender_registry_key_replace_hkcr",
        transformation=ReplaceStringTransformation(regex=r"(?i)(^HKCR)",
                                                   replacement=r"HKEY_LOCAL_MACHINE\\CLASSES"),
        field_name_conditions=[IncludeFieldCondition(['RegistryKey', 'PreviousRegistryKey'])]
    ),
    ProcessingItem(
        identifier="microsoft_365_defender_registry_actiontype_value",
        transformation=RegistryActionTypeValueTransformation(),
        field_name_conditions=[IncludeFieldCondition(['ActionType'])]
    ),
    # Extract Domain from Username fields
    ProcessingItem(
        identifier="microsoft_365_defender_domain_username_extract",
        transformation=SplitDomainUserTransformation(),
        field_name_conditions=[IncludeFieldCondition(["AccountName", "InitiatingProcessAccountName"])]
    ),
    ProcessingItem(
        identifier="microsoft_365_defender_hashes_field_values",
        transformation=HashesValuesTransformation(),
        field_name_conditions=[IncludeFieldCondition(['Hashes'])]
    ),
]

## Exceptions/Errors ProcessingItems
rule_error_proc_items = [
    # Category Not Supported
    ProcessingItem(
        identifier="microsoft_365_defender_unsupported_rule_category",
        rule_condition_linking=any,
        transformation=RuleFailureTransformation(
            "Rule category not yet supported by the Microsoft 365 Defender Sigma backend."
        ),
        rule_condition_negation=True,
        rule_conditions=[x for x in category_to_conditions_mappings.values()],
    )]

field_error_proc_items = [
    # Invalid fields per category
    ProcessingItem(
        identifier=f"microsoft_365_defender_unsupported_fields_{table_name}",
        transformation=InvalidFieldTransformation(
            f"Please use valid fields for the {table_name} table, or the following fields that have keymappings in this "
            f"pipeline:\n"
            # Combine field mappings for table and generic field mappings dicts, get the unique keys, add the Hashes field, sort it
            f"{', '.join(sorted(set({**query_table_field_mappings[table_name], **generic_field_mappings}.keys()).union({'Hashes'})))}"
        ),
        field_name_conditions=[
            ExcludeFieldCondition(fields=table_fields + list(generic_field_mappings.keys()) + ['Hashes'])],
        rule_conditions=[
            category_to_conditions_mappings[rule_category]
            for rule_category in table_to_category_mappings[table_name]
        ],
        rule_condition_linking=any,
    )
    for table_name, table_fields in valid_fields_per_table.items()
]


def microsoft_365_defender_pipeline() -> ProcessingPipeline:
    """Pipeline for transformations for SigmaRules to use in the Microsoft 365 Defender Backend
    Field mappings based on documentation found here:
    https://learn.microsoft.com/en-us/microsoft-365/security/defender/advanced-hunting-query-language?view=o365-worldwide
    :return: ProcessingPipeline for Microsoft 365 Defender Backend
    :rtype: ProcessingPipeline
    """
    return ProcessingPipeline(
        name="Generic Log Sources to Windows 365 Defender Transformation",
        priority=10,
        items=[*query_table_proc_items,
               *fieldmappings_proc_items,
               *generic_field_mappings_proc_item,
               *replacement_proc_items,
               *rule_error_proc_items,
               *field_error_proc_items,
               ]
    )
